#ifndef _MISC_H_SEEN_
#define _MISC_H_SEEN_

struct cmdarg
{
	char *str;
	struct cmdarg *n;
};

struct cmdarg *add_cmdarg(struct cmdarg *, const char *str);
char *cat_args(struct cmdarg *);

#endif /* _MISC_H_SEEN_ */
