/* Common headers and defns */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <assert.h>

#ifdef __cplusplus
#define EXTERNC extern "C"
#endif
