/*
 * Inode data for userfs
 */

#ifndef _LINUX_USERFS_FS_I
#define _LINUX_USERFS_FS_I

/*
 * What a handle looks like
 */
typedef struct
{
  unsigned long handle;
} up_handle;

struct userfs_inode_info
{
  up_handle	handle;
  struct userfs_dir_ra	*dir_ra;
};

#endif /* _LINUX_USERFS_FS_I */
