/* -*- C -*- */

#ifndef _LIB_IO_H_
#define _LIB_IO_H_

/*
 * Simple IO functions
 */

#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif
	
int fullread(int fd, unsigned char *, size_t);
int fullwrite(int fd, const unsigned char *, size_t);

#ifdef __cplusplus
}
#endif

#endif /* _LIB_IO_H_ */
