/*
 * Per file open information for userfs
 */

#ifndef __LINUX_USERFS_FS_F_H_SEEN__
#define __LINUX_USERFS_FS_F_H_SEEN__

typedef unsigned long up_credtok;

struct userfs_file_info {
	up_credtok cred_tok;		/* credentials token */
};

#endif /* __LINUX_USERFS_FS_F_H_SEEN__ */
