
const char *opnames[] =
{
	"up_create",
	"up_lookup",
	"up_close",
	"up_read",
	"up_write",
	"up_truncate",
	"up_fsync",
	"up_readdir",
	"up_link",
	"up_unlink",
	"up_symlink",
	"up_readlink",
	"up_followlink",
	"up_mount",
	"up_umount",
	"up_iread",
	"up_iwrite",
	"up_statfs",
	"up_iput",
	"up_open",
	"up_permission",
	"up_rename",
};

